#!/usr/bin/env python

"""
RSVP instruction and serialisation classes.

Copyright (C) 2007, 2008, 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.data import Attr, Const, Class, Function, Module
from micropython.program import Block, DataObject, DataValue, PlaceholderContext
from micropython.raw import RawObject

def name(attr):
    if isinstance(attr, Attr):
        return attr.name or "<unnamed>"
    elif isinstance(attr, (Block, Const)):
        return attr
    else:
        return attr.full_name() or "<unnamed>"

# Serialisation-related classes.

class RSVPObject(RawObject):

    "A generic data object wrapper."

    def _finalise_location(self, with_builtins):

        """
        Set the code body location for items now that the code blocks have been
        positioned.
        """

        item = self.item

        if not self.is_generated(with_builtins):
            item.code_body_location = item.full_name()
        else:
            item.code_body_location = item.get_body_block().location

class RSVPAttr(RSVPObject):

    "A wrapper for attributes/values."

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        return [
            DataValue(
                item.get_context() and item.get_context().location,
                item.get_value() and item.get_value().location
                )
            ]

class RSVPBlock(RSVPObject):

    "A wrapper for blocks."

    def set_location(self, location, objtable, with_builtins):
        item = self.item
        item.location = location
        return location + len(item.code)

    def as_raw(self, objtable, paramtable, with_builtins):
        return self.item.code

class RSVPClass(RSVPObject):

    "A wrapper for classes."

    def set_location(self, location, objtable, with_builtins):
        self.item.instance_template_location = location

        # Include the instance template and __class__ attribute in the size.

        return RSVPObject.set_location(self, location + 1, objtable, with_builtins)

    def finalise_location(self, with_builtins):
        self._finalise_location(with_builtins)

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        classcode = objtable.as_list().get_code(item.full_name())
        attrcode = objtable.get_index(item.full_name())

        # Include a template of an instance for use when instantiating classes.

        call_method = item.all_class_attributes().get("__call__")
        call_method_value = call_method and call_method.get_value()
        call_method_code_location = call_method_value and call_method_value.code_location
        call_method_funccode = call_method_value and paramtable.as_list().get_code(call_method_value.full_name())

        instantiator_funccode = paramtable.as_list().get_code(item.get_instantiator().full_name())

        # NOTE: The instantiator code is the first block of the class.

        if not self.is_generated(with_builtins):
            instantiator_code_location = item.full_name()
        else:
            instantiator_code_location = item.get_instantiator().blocks[0].location

        return [

            # Template instance...

            DataObject(
                classcode,
                attrcode, # is instance
                call_method_code_location,
                item.full_name(),
                len(item.instance_attributes()), # size (excluding __class__)
                call_method_funccode # funccode
                ),

            # Class...

            DataObject(
                classcode,
                None, # is not instance
                instantiator_code_location,
                item.full_name(),
                len(item.class_attributes()) + 1, # size
                instantiator_funccode # funccode
                )
            ]

class RSVPConst(RSVPObject):

    "A wrapper for constants."

    def set_location(self, location, objtable, with_builtins):
        item = self.item
        value = item.get_value()
        type_name = self.type_name(item.value_type_name(), value, objtable)

        location = RSVPObject.set_location(self, location, objtable, with_builtins)
        return location + self.raw_size(type_name, value)

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        value = item.get_value()
        type_name = self.type_name(item.value_type_name(), value, objtable)
        raw_data = self.raw_data(type_name, value, objtable)
        size = self.raw_size(type_name, value) + 1

        # Generate the footprint of the constant.

        classcode = objtable.as_list().get_code(type_name)
        attrcode = objtable.get_index(type_name)

        return [
            DataObject(
                classcode,
                attrcode, # is instance
                None,
                type_name,
                size # header plus data
                )

            # NOTE: The RSVP library needs changing if more attributes are added
            # NOTE: here (such as a __class__ attribute for all instances).

            ] + raw_data

    def type_name(self, type_name, value, objtable):

        # Change the type of string constants if they might be used as attribute
        # accessors.

        accessor_cls = "__builtins__._accessor"

        if type_name == "__builtins__.str" and value in objtable.attribute_names() and \
            accessor_cls in objtable.object_names():

            return accessor_cls
        else:
            return type_name

    def raw_data(self, type_name, value, objtable):

        # NOTE: Start simple and use single entries for most types, even though
        # NOTE: a concrete representation may use multiple memory words to
        # NOTE: represent the data.

        if type_name in ("__builtins__.tuple", "__builtins__.list"):
            return [len(value)] + list(value)
        elif type_name == "__builtins__._accessor":
            return [value, objtable.get_index(value)]
        else:
            return [value]

    def raw_size(self, type_name, value):
        if type_name in ("__builtins__.tuple", "__builtins__.list"):
            return 1 + len(value)
        elif type_name == "__builtins__._accessor":
            return 2
        else:
            return 1

class RSVPFunction(RSVPObject):

    "A wrapper for functions."

    def set_location(self, location, objtable, with_builtins):
        item = self.item
        location = RSVPObject.set_location(self, location, objtable, with_builtins)

        # Set the code location only where the code has been
        # generated.

        if not self.is_generated(with_builtins):
            item.code_location = item.full_name()

        # Skip any defaults for static functions.

        elif not item.is_dynamic():
            item.code_location = location + len(item.defaults)

        # Skip any defaults for dynamic functions.

        else:
            item.code_location = location

        return location

    def finalise_location(self, with_builtins):
        self._finalise_location(with_builtins)

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        # NOTE: Need class and parameter details! Should arguably be an instance of types.FunctionType.
        return [
            DataObject(
                objtable.as_list().get_code("__builtins__.function"),
                objtable.get_index("__builtins__.function"), # is instance
                item.code_location,
                "__builtins__.function",
                len(item.defaults) + 1, # size (not accurate for lambda functions before instantiation)
                paramtable.as_list().get_code(item.full_name()) # funccode
                )
            ]

class RSVPModule(RSVPObject):

    "A wrapper for modules."

    def as_raw(self, objtable, paramtable, with_builtins):
        item = self.item
        return [
            DataObject(
                objtable.as_list().get_code(item.full_name()),
                None, # modules treated like classes
                None,
                item.full_name(),
                len(item.module_attributes()) + 1 # size
                )
            ]

# Serialisation-related data and functions.

def get_object(item):
    if isinstance(item, Attr):
        cls = RSVPAttr
    elif isinstance(item, Block):
        cls = RSVPBlock
    elif isinstance(item, Class):
        cls = RSVPClass
    elif isinstance(item, Const):
        cls = RSVPConst
    elif isinstance(item, Function):
        cls = RSVPFunction
    elif isinstance(item, Module):
        cls = RSVPModule
    else:
        cls = None

    if cls is not None:
        return cls(item)
    else:
        return None

# Instruction-related classes.

class Instruction:

    "A generic instruction."

    # NOTE: Ultimately, instructions apart from Transfer will use specific
    # NOTE: registers such as "working_value" and "working_context".

    def __init__(self, attr=None, working="working", target="working", source=None):
        self.attr = attr
        self.working = working
        self.target = target
        self.source = source

    def get_details(self):
        return self.__class__, self.attr, self.working, self.target, self.source

    def copy(self):
        return self.__class__(self.attr, self.working, self.target, self.source)

    def __repr__(self):
        return "%s(%s%s%s%s)" % (self.__class__.__name__,
            self.format_operand(),
            self.format_working(),
            self.format_source(),
            self.format_target())

    def __hash__(self):
        return hash(self.get_details())

    def __eq__(self, other):
        return self.get_details() == other.get_details()

    def __ne__(self, other):
        return not self.__eq__(other)

    def format_operand(self):
        operand = self.get_operand()
        return repr(operand)

    def format_working(self):
        return self.working != "working" and (", working=%r" % self.working) or ""

    def format_source(self):
        return self.source is not None and (", source=%r" % self.source) or ""

    def format_target(self):
        return self.target != "working" and (", target=%r" % self.target) or ""

    def get_operand(self):
        return None

class FrameRelativeInstruction(Instruction):

    "An instruction operating on the current frame."

    def get_operand(self):
        return self.attr.position

FR = FrameRelativeInstruction

class AddressRelativeInstruction(Instruction):

    "An instruction accessing an object's attribute."

    def format_operand(self):
        return "%s, name=%r" % (Instruction.format_operand(self), name(self.attr))

    def get_operand(self):
        return self.attr.position

AR = AddressRelativeInstruction

class AddressInstruction(Instruction):

    "An instruction loading an address directly."

    def format_operand(self):
        location, position, result = self.get_operands()
        if location is not None:
            return "%s, location=%s, position=%s, name=%s" % (
                result, location, position, name(self.attr))
        elif result is not None:
            return "%s, name=%s" % (result, name(self.attr))
        else:
            return "name=%s" % name(self.attr)

    def get_operands(self):
        if isinstance(self.attr, Attr):
            position = self.attr.position
            location = self.attr.parent.location

            # NOTE: Unpositioned attributes are handled here.

            if location is not None and position is not None:
                result = location + position + 1
            else:
                location = self.attr.parent.name
                position = self.attr.name
                result = None
            return location, position, result
        else:
            return None, None, self.attr.location

    def get_operand(self):
        return self.get_operands()[-1]

Address = AddressInstruction

class TargetInstruction(Instruction):

    "An instruction loading the address of an invocation target."

    def format_operand(self):
        return "%s, name=%r" % (Instruction.format_operand(self), name(self.attr))

    def get_operand(self):
        return self.attr.code_body_location

Target = TargetInstruction

class ImmediateInstruction(Instruction):

    "An instruction employing a constant."

    def get_operand(self):
        return self.attr

Immediate = ImmediateInstruction

# Low-level instructions.

class Transfer(Instruction):
    "Transfer a register's contents into another."
    cost = 1

# Access to stored constant data.

class LoadConst(Address):
    "Load the constant or module from the specified location."
    cost = 1

class LoadClass(Address):
    "Load the class from the specified location."
    cost = 1

class LoadFunction(Address):
    "Load the function from the specified location."
    cost = 1

# Access within an invocation frame.

class LoadName(FR):
    "Load the current value from the given local attribute/variable."
    cost = 2

class StoreName(FR):
    "Store the source value into the given local attribute/variable."
    cost = 2

class LoadTemp(Immediate):
    "Load the current value from the given temporary location."
    cost = 2

class StoreTemp(Immediate):
    "Store the current value into the given temporary location."
    cost = 2

# Access to static data.

class LoadAddress(Address):
    "Load the current value from the given fixed attribute address."
    cost = 1

class StoreAddress(Address):
    "Store the source value into the given fixed attribute address."
    cost = 1

class LoadAddressContext(Address):
    "Load the current value from the given fixed attribute address, using the current value as context."
    cost = 2

class StoreAddressContext(Address):
    "Store the current value into the given fixed attribute address, using the current value as context."
    cost = 2

class LoadAddressContextCond(Address):
    """
    Load the current value from the given fixed attribute address, only using the current value as
    context if the attribute is compatible.
    """
    cost = 4

class MakeInstance(Immediate):
    "Make a new instance using the current value as a reference to a template."
    cost = 5

class MakeFragment(Immediate):
    "Make a new list fragment."
    cost = 5

# Access to address-relative data. (LoadAttrIndexContext not defined.)

class LoadAttr(AR):
    "Load into the current value the given attribute of the object referenced by the current value."
    cost = 2

class StoreAttr(AR):
    "Store the source value into the given attribute of the object referenced by the current value."
    cost = 2

class LoadAttrIndex(Immediate):
    "Load into the current value the attribute of the current value with the given index."
    cost = 6

class StoreAttrIndex(Immediate):
    "Store the source value into the attribute of the current value with the given index."
    cost = 6

class LoadAttrIndexContextCond(Immediate):
    """
    Load into the current value the attribute of the current value with the given index, only making the
    current value the context if the attribute is compatible.
    """
    cost = 8

# Access to object details.

class LoadCallable(Instruction):
    "Load the target of an invocation."
    cost = 2

class StoreCallable(Instruction):
    "Store the source value into the object referenced by the current value."
    cost = 3

# Access to invocation frames in preparation.

class MakeFrame(Immediate):
    "Make a new invocation frame."
    cost = 2

class AdjustFrame(Immediate):
    "Adjust the current invocation frame for corrected invocations."
    cost = 2

class DropFrame(Instruction):
    "Drop an invocation frame."
    cost = 2

class StoreFrame(Immediate):
    "Store the current value as an argument for the parameter with the given position."
    cost = 2

class StoreFrameIndex(Immediate):
    "Store the source value as an argument of the current value for the parameter with the given index."
    cost = 6

# Context-related tests.

class CheckContext(Instruction):
    "Check to see if the context is valid."
    cost = 2

class CheckClass(Instruction):
    "Check the current value to determine whether it is a class."
    cost = 2

class CheckInstance(Instruction):
    """
    Check the current value as an instance of a class or its subclasses (used with 'self' in an
    invocation).
    """
    cost = 6

# Access to frames upon invocation.

class CheckFrame(Immediate):
    "Check the frame for the correct number of arguments."
    cost = 3

class CheckExtra(Immediate):
    "Ensure that the frame can provide extra arguments."
    cost = 3

class FillDefaults(Immediate):
    "Fill frame positions with defaults, if appropriate."
    cost = 8 # variable

class ExtendFrame(Immediate):
    "Extend the current frame for temporary storage use."
    cost = 1

class CopyExtra(Immediate):
    "Copy extra arguments into a separate sequence, starting from the given position."
    cost = 10

# Invocation-related instructions, using a special result "register".

class JumpInFrame(Instruction):
    "Jump, using the current locals, to the current callable."
    cost = 2

class JumpWithFrame(Instruction):
    "Jump, adopting the invocation frame, to the current callable."
    cost = 3

class JumpWithFrameDirect(Target):
    "Jump to the specified address, adopting the invocation frame."
    cost = 3

class Return(Instruction):
    "Return from a subprogram."
    cost = 2

# Branch-related instructions.

class Jump(Address):
    "Jump unconditionally."
    cost = 1

class JumpIfFalse(Address):
    "Jump if the last evaluation gave a false result."
    cost = 2

class JumpIfTrue(Address):
    "Jump if the last evaluation gave a true result."
    cost = 2

# Exception-related instructions, using a special exception "register".

class RaiseException(Instruction):
    "Raise an exception, jumping to the active handler."
    cost = 2

class PushHandler(Address):
    "Push an exception handler onto the handler stack."
    cost = 3

class PopHandler(Immediate):
    "Pop exception handlers from the handler stack."
    cost = 3

class CheckException(Instruction):
    "Check the raised exception against another."
    cost = 6

class ClearException(Instruction):
    "Clear any raised exception."
    cost = 1

# Test instructions, operating on the boolean status register.

class TestIdentity(Instruction):
    "Test whether the current value is identical to the source value, setting the boolean status."
    cost = 2

class TestIdentityAddress(Address):
    "Test whether the current value is identical to the given address, setting the boolean status."
    cost = 2

class InvertBoolean(Instruction):
    "Invert the boolean status."
    cost = 1

# Instructions which can affect the current value. (LoadAttrIndexContext not defined.)

current_value_instructions = (
    Transfer,
    LoadConst, LoadClass, LoadFunction,
    LoadName, LoadTemp,
    LoadAddress, LoadAddressContext, LoadAddressContextCond,
    LoadAttr, LoadAttrIndex, LoadAttrIndexContextCond,
    LoadCallable,
    MakeInstance, MakeFragment,
    CopyExtra,
    CheckClass, CheckContext, CheckException, CheckInstance, CheckFrame, CheckExtra
    )

# vim: tabstop=4 expandtab shiftwidth=4
